# -*- coding: utf-8 -*-
# Copyright (c) 2022 Aleksandar Buza <me@aleksandarbuza.com>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import sys
from pathlib import Path
from typing import List, Tuple

from punsctl.namespace import Namespace, NamespaceException
from punsctl.rootspace import RootSpace, RootSpaceException
from punsctl.sgetopt import sgetopt
from punsctl.static import DEFAULT_ROOT_PATH, DEFAULT_SYMLINK_PATH, USAGE


@sgetopt(args=sys.argv[1:], optstring="hlxr:s:n:d:a:")
def main(opts: List[Tuple], argv: List[str]) -> None:
    if len(opts) == 0:
        sys.exit(USAGE)

    opt_list = False
    opt_root_path = DEFAULT_ROOT_PATH
    opt_symlink_path = DEFAULT_SYMLINK_PATH
    opt_create = None
    opt_delete = None
    opt_activate = None
    opt_deactivate = False

    for opt, arg in opts:
        if opt == "-h":
            sys.exit(USAGE)

        elif opt == "-l":
            opt_list = True

        elif opt == "-x":
            opt_deactivate = True

        elif opt == "-r":
            opt_root_path = arg if arg is not None else sys.exit(USAGE)

        elif opt == "-s":
            opt_symlink_path = arg if arg is not None else sys.exit(USAGE)

        elif opt == "-n":
            opt_create = arg if arg is not None else sys.exit(USAGE)

        elif opt == "-d":
            opt_delete = arg if arg is not None else sys.exit(USAGE)

        elif opt == "-a":
            opt_activate = arg if arg is not None else sys.exit(USAGE)

        else:
            sys.exit(USAGE)

    try:
        root_space = RootSpace(
            path=Path(opt_root_path), symlink_path=Path(opt_symlink_path)
        )

    except RootSpaceException as exc:
        sys.exit(f"root error: {exc.message}")

    if opt_list:
        for namespace in root_space.get_all_ns_paths():
            try:
                ns = Namespace(
                    name=namespace.name,
                    root_space=root_space,
                )

                sys.stdout.write(
                    f"{namespace.name} ({namespace.absolute()}) "
                    f"{'active' if ns.active() else ''}\n"
                )
            except NamespaceException as exc:
                sys.exit(f"namespace error: {exc.message}")

    elif opt_create is not None:
        ns = Namespace(name=opt_create, root_space=root_space)

        try:
            ns.create()
            sys.stdout.write(f"info: {opt_create} created\n")

        except NamespaceException as exc:
            sys.exit(f"namespace error: {exc.message}")

    elif opt_delete is not None:
        ns = Namespace(name=opt_delete, root_space=root_space)

        try:
            ns.remove()
            sys.stdout.write(f"info: {opt_delete} removed\n")

        except NamespaceException as exc:
            sys.exit(f"namespace error: {exc.message}")

    elif opt_activate is not None:
        ns = Namespace(name=opt_activate, root_space=root_space)

        try:
            ns.activate()
            sys.stdout.write(f"info: {opt_activate} activated\n")

        except NamespaceException as exc:
            sys.exit(f"namespace error: {exc.message}")

    elif opt_deactivate:
        for namespace in root_space.get_all_ns_paths():
            ns = Namespace(name=namespace.name, root_space=root_space)
            ns.deactivate()

        sys.stdout.write("info: namespaces are deactivated successfully\n")

    else:
        sys.stdout.write(USAGE)



# -*- coding: utf-8 -*-
# Copyright (c) 2022 Aleksandar Buza <me@aleksandarbuza.com>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from os import mkdir, readlink, rmdir
from pathlib import Path
from typing import List

from punsctl.rootspace import RootSpace

__all__ = ["Namespace", "NamespaceException"]


class NamespaceException(Exception):
    def __init__(self, message):
        self.message = message


class Namespace(object):
    def __init__(
        self,
        name: str,
        root_space: RootSpace,
    ):
        self.name = name
        self.root_space = root_space

        self.symlink_path = root_space.get_symlink_path()
        self.current_ns_path = root_space.get_current_ns_path()
        self.path = Path(f"{root_space.get_path()}/{name}")

    def create(self) -> None:
        try:
            mkdir(self.path, mode=0o744)

        except OSError as exc:
            raise NamespaceException(message=exc.strerror)

    def remove(self) -> None:
        try:
            if self.exists():
                rmdir(self.path)

        except OSError as exc:
            raise NamespaceException(message=exc.strerror)

    def exists(self) -> bool:
        if not Path(self.path).exists():
            return False

        return True

    def get_path(self) -> Path:
        return self.path

    def active(self) -> bool:
        if (
            self.current_ns_path.exists()
            and Path(readlink(self.current_ns_path)) == self.path
        ):
            return True

        return False

    def get_name(self) -> str:
        return self.name

    def get_namespace_path(self) -> Path:
        return self.path

    def __get_sources(self) -> List[Path]:
        sources = []

        for source in self.path.iterdir():
            if source.name == self.current_ns_path.name:
                continue
            sources.append(source)

        return sources

    def activate(self) -> None:
        if self.current_ns_path.exists() and self.current_ns_path.is_symlink():
            if Path(readlink(self.current_ns_path)) == self.path:
                pass

            else:
                raise NamespaceException(
                    message=(
                        f"{self.root_space.get_current_ns_name()} "
                        f"namespace are already activated"
                    )
                )
        else:
            self.current_ns_path.symlink_to(self.path)

        for source in self.__get_sources():
            with Path(f"{self.symlink_path}/{source.name}") as target:
                if target.exists():
                    if target.is_symlink() and Path(readlink(target)) == source:
                        continue

                    with target.with_name(f"{target.name}.{self.name}.bak") as backup:
                        if not target.is_symlink():
                            target.rename(backup)

                if not target.is_symlink():
                    target.symlink_to(source)

    def deactivate(self) -> None:
        if self.current_ns_path.exists() and self.current_ns_path.is_symlink():
            if Path(readlink(self.current_ns_path)) == self.path:
                self.current_ns_path.unlink()

        for source in self.__get_sources():
            with Path(f"{self.symlink_path}/{source.name}") as target:
                if target.exists():
                    if target.is_symlink() and Path(readlink(target)) == source:
                        target.unlink()

                    with target.with_name(f"{target.name}.{self.name}.bak") as backup:
                        if backup.exists():
                            backup.rename(target)



# -*- coding: utf-8 -*-
# Copyright (c) 2022 Aleksandar Buza <me@aleksandarbuza.com>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import Any, List

__all__ = ["sgetopt"]


def sgetopt(args: List[str], optstring: str) -> Any:
    """
    Parses command line options and parameter list.
    args is the argument list to be parsed, without the leading reference
    to the running program. Typically, this means "sys.argv[1:]".
    optstring is the string of option letters that the script wants to
    recognize, with options that require an argument followed by a
    colon (i.e., the same format that Unix getopt() uses).
    """

    def func(f):
        def wrapper():
            def is_opt(opt: str) -> bool:
                """
                Check if the argument is an option
                """
                if len(opt) == 2 and opt.startswith("-"):
                    return True
                return False

            def req_arg(opt: str) -> bool:
                """
                Check if the option requires an argument
                """
                if opt.isalpha():
                    return optstring[optstring.find(opt) + 1] == ":"
                return False

            opts, argv = [], []

            for _ in range(len(args)):
                arg, arg_ix = args[_], _

                if is_opt(arg):
                    n_arg = arg_ix + 1

                    if len(args) <= n_arg:
                        opts.append((arg, None))

                    elif req_arg(arg[1]) and not is_opt(args[n_arg]):
                        opts.append((arg, args[n_arg]))

                    else:
                        opts.append((arg, ""))

                else:
                    p_arg = arg_ix - 1

                    if is_opt(args[p_arg]):
                        if not req_arg(args[p_arg][1]):
                            argv.append(arg)

                    else:
                        argv.append(arg)

            f(opts, argv)

        return wrapper

    return func



# -*- coding: utf-8 -*-
# Copyright (c) 2022 Aleksandar Buza <me@aleksandarbuza.com>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from os import R_OK, W_OK, access, readlink
from pathlib import Path
from typing import List, Optional

from punsctl.static import DEFAULT_ROOT_PATH, DEFAULT_SYMLINK_PATH

__all__ = ["RootSpace", "RootSpaceException"]


class RootSpaceException(Exception):
    def __init__(self, message):
        self.message = message


class RootSpace(object):
    def __init__(
        self,
        path: Path = Path(DEFAULT_ROOT_PATH),
        symlink_path: Path = Path(DEFAULT_SYMLINK_PATH),
    ):
        # Root path checks
        if not path.exists():
            raise RootSpaceException(message=f"path {path} doesn't exists")

        if not path.is_dir():
            raise RootSpaceException(message=f"path {path} is not a directory")

        if access(path, R_OK) is not True:
            raise RootSpaceException(message=f"path {path} is not readable")

        # Symlink path checks
        if not symlink_path.exists():
            raise RootSpaceException(message=f"path {symlink_path} doesn't exists")

        if not symlink_path.is_dir():
            raise RootSpaceException(message=f"path {symlink_path} is not a directory")

        if access(symlink_path, W_OK) is not True:
            raise RootSpaceException(message=f"path {symlink_path} is not writable")

        self.path = path
        self.symlink_path = symlink_path
        self.current_ns_path = Path(f"{symlink_path}/.current_ns")

    def get_path(self) -> Path:
        return self.path

    def get_symlink_path(self) -> Path:
        return self.symlink_path

    def get_current_ns_path(self) -> Path:
        return self.current_ns_path

    def get_current_ns_name(self) -> Optional[str]:
        if self.current_ns_path.exists() and self.current_ns_path.is_symlink():
            return Path(readlink(self.current_ns_path)).name

        return None

    def get_all_ns_paths(self) -> List[Path]:
        namespaces = []

        for namespace in self.path.iterdir():
            if namespace.is_dir() and not namespace.is_symlink():
                namespaces.append(namespace)

        namespaces.sort()

        return namespaces



# -*- coding: utf-8 -*-
# Copyright (c) 2022 Aleksandar Buza <me@aleksandarbuza.com>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from pathlib import Path

USAGE = """
punsctl <options>

options:
    -h                  Help menu
    -r                  Root path                 (Default: ~/.ns)
    -s                  Symlink path              (Default: ~/)
    -l                  List namespaces
    -n <namespace>      Create a new namespace
    -d <namespace>      Delete existing namespace
    -a <namespace>      Activate namespace
    -x                  Deactivate namespaces
"""

DEFAULT_ROOT_PATH = f"{Path.home()}/.ns"
DEFAULT_SYMLINK_PATH = f"{Path.home()}"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['punsctl']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['punsctl = punsctl.main:main']}

setup_kwargs = {
    'name': 'punsctl',
    'version': '0.1.0',
    'description': "POSIX User's Namespace Control",
    'long_description': '### `The project is currently under development and is not ready for use in production.`\n\n# punsctl - POSIX User\'s Namespace Control\n\n[![codecov](https://codecov.io/github/alekbuza/punsctl/branch/main/graph/badge.svg?token=OMHOSME5ZB)](https://codecov.io/github/alekbuza/punsctl)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/punsctl)\n![PyPI - Implementation](https://img.shields.io/pypi/implementation/punsctl)\n![PyPI - Wheel](https://img.shields.io/pypi/wheel/punsctl)\n\nThe `punsctl` utility manages multiple namespaces (user environments) for the current POSIX user.\nThat means the user can have multiple "_profiles_" for the same or different tool configurations in the same user account\n(`~/.ssh`, `~/.gitconfig`, `~/.gnupg`, `~/.config`, `~/.config/nvim`, ...).\nThe user can create, delete, activate and deactivate namespaces without additional permissions.\n\n## Installation\n\n```sh\npip install punsctl\n```\n\n## Usage\n\n### Create a new namespace\n```sh\npunsctl -c <namespace>\n```\n\n### List all namespaces\n```sh\npunsctl -l\n```\n\n### Activate a namespace\n```sh\npunsctl -a <namespace>\n```\n\n### Deactivate namespaces\n```sh\npunsctl -d\n```\n\n### List all namespaces for the non-default root path (`~/.ns`)\n```sh\npunsctl -p <root_path> -l\n```\n',
    'author': 'Aleksandar Buza',
    'author_email': 'me@aleksandarbuza.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/alekbuza/punsctl',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
